/*
 * menu.c - menu dialogs.
 *
 * $Id: menu.c 30 2006-02-09 15:29:48Z brendan $
 *
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 * 29-Jan-2006  Brendan Gregg   Created this
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <strings.h>
#include <unistd.h>
#include <libgen.h>
#include <gtk/gtk.h>

#include "dtaz.h"
#include "colours.h"
#include "dialog.h"
#include "menu.h"
#include "plot.h"
#include "system.h"
#include "statusbar.h"

/* private variables */
static char	lastname[LINE_LEN];

/* private functions */
static void menu_save_dialog(gpointer, guint, GtkWidget *);
static void menu_open_dialog(gpointer, guint, GtkWidget *);
static void menu_save_file(char *);
static void menu_open_file(char *);
static void menu_set_pallet(gchar *);

/*
 * Menu messages
 */
static const gchar	*menu_options_help_msg = "Resolution\n\n"
"	This is the width and height of the blocks drawn on the main plot.\n"
"	The units are pixels.\n\n\n"
"Colour Factor\n\n"
"	This is the factor used to calculate the colour drawn in the disk\n"
"	activity block for the main plot.\n\n"
"	If your plot is too \"red\", increase this value.\n"
"	If your plot is too \"blue\", decrease this value.\n\n\n"
"Update Time\n\n"
"	This is the time in milliseconds between updating the plots.\n\n";

/*
 * Menu constructor
 *
 * BDG - I should stop using GtkItemFactory as it's deprecated.
 */
static GtkItemFactoryEntry menu_items[] = {
	{ "/_File",	NULL,	NULL, 0, "<Branch>" },
	{ "/File/_New",	"<control>N",	plot_new_screen, 0, NULL },
	{ "/File/_Open",	"<control>O",	menu_open_dialog, 0, NULL },
	{ "/File/_Save",	"<control>S",	menu_save_dialog, 0, NULL },
	{ "/File/Save _As",	NULL,	menu_save_dialog, 0, NULL },
	{ "/File/sep1",	NULL,	NULL,	0,	"<Separator>" },
	{ "/File/Quit",	"<control>Q",	die_blocking, 0, NULL },
	{ "/_Options",	NULL,	NULL,	0, "<Branch>" },
	{ "/Options/Options",	NULL,	menu_options, 1, NULL },
	{ "/_Help",	NULL,	NULL,	0, "<LastBranch>" },
	{ "/_Help/About",	NULL,	dialog_about, 1, NULL },
};

void
menu_create_main(GtkWidget *window, GtkWidget **menubar)
{
	GtkItemFactory	*item_factory;
	GtkAccelGroup	*accel_group;
	gint		nmenu_items;

	nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

	/* shortcuts */
	accel_group = gtk_accel_group_new();

	/* menu creater */
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>",
	    accel_group);

	/* generate menu items */
	gtk_item_factory_create_items(item_factory, nmenu_items,
	    menu_items, NULL);

	gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);

	if (menubar)
		*menubar = gtk_item_factory_get_widget(item_factory, "<main>");
}

void
menu_options_help(GtkWidget *window)
{
	GtkWidget	*dialog;
	GtkWidget	*view;
	GtkTextBuffer	*buffer;

	/* create the widgets */
	dialog = gtk_dialog_new_with_buttons("Options Help",
	    GTK_WINDOW(window),
	    GTK_DIALOG_DESTROY_WITH_PARENT,
	    GTK_STOCK_OK,
	    GTK_RESPONSE_NONE,
	    NULL);

	/* create text box */
	view = gtk_text_view_new();
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
	gtk_text_buffer_set_text(GTK_TEXT_BUFFER(buffer),
	    menu_options_help_msg, -1);

	/* add view */
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), view);

	/* show, wait, destroy */
	gtk_widget_show_all(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

void
menu_set_pallet(gchar *selected)
{
	/* set global pallet */
	if (strncmp(selected, "greyscale", 4) == 0)
		c_pallet = C_GREYSCALE;
	else if (strncmp(selected, "scale12", 4) == 0)
		c_pallet = C_SCALE12;
	else
		c_pallet = C_ORIGINAL;

	/* realloc colours */
	colours_set(GTK_WINDOW_TOPLEVEL, FALSE);
}

void
menu_options(gpointer data, guint action, GtkWidget *menuitem)
{
	GtkWidget	*dialog;
	GtkWidget	*table;
	GtkWidget	*label[4];
	GtkWidget	*scale[3];
	GtkWidget	*seperator[3];
	GtkObject	*adjustment[3];
	GtkWidget	*combo;
	GList		*glist = NULL;
	gint		result;
	gint		running;

	/* create the widgets */
	dialog = gtk_dialog_new_with_buttons("Options",
	    GTK_WINDOW_TOPLEVEL,
	    GTK_DIALOG_DESTROY_WITH_PARENT,
	    GTK_STOCK_OK,
	    GTK_RESPONSE_ACCEPT,
	    GTK_STOCK_APPLY,
	    GTK_RESPONSE_APPLY,
	    GTK_STOCK_CANCEL,
	    GTK_RESPONSE_REJECT,
	    GTK_STOCK_HELP,
	    GTK_RESPONSE_HELP,
	    NULL);

	/* create table for layout */
	table = gtk_table_new(1, 11, FALSE);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), table);

	/* add labels */
	label[0] = gtk_label_new("Resolution (pixels)");
	label[1] = gtk_label_new("Colour Factor");
	label[2] = gtk_label_new("Colour Pallet");
	label[3] = gtk_label_new("Update Time (ms)");
	gtk_table_attach(GTK_TABLE(table), label[0],
	    0, 1, 0, 1, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 5, 2);
	gtk_table_attach(GTK_TABLE(table), label[1],
	    0, 1, 3, 4, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 5, 2);
	gtk_table_attach(GTK_TABLE(table), label[2],
	    0, 1, 6, 7, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 5, 2);
	gtk_table_attach(GTK_TABLE(table), label[3],
	    0, 1, 9, 10, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 5, 2);

	/* add horizontal scales */
	adjustment[0] = gtk_adjustment_new(dt_resolution,
	    1, MENU_RESOLUTION_MAX, 1, 2, 0);
	adjustment[1] = gtk_adjustment_new(dt_colour_factor,
	    1, MENU_COLOUR_FACTOR_MAX, 1, 4, 0);
	adjustment[2] = gtk_adjustment_new(dt_update,
	    100, MENU_UPDATE_TIME_MAX, 100, 100, 0);
	scale[0] = gtk_hscale_new(GTK_ADJUSTMENT(adjustment[0]));
	scale[1] = gtk_hscale_new(GTK_ADJUSTMENT(adjustment[1]));
	scale[2] = gtk_hscale_new(GTK_ADJUSTMENT(adjustment[2]));
	gtk_scale_set_digits(GTK_SCALE(scale[0]), 0);
	gtk_scale_set_digits(GTK_SCALE(scale[1]), 0);
	gtk_scale_set_digits(GTK_SCALE(scale[2]), 0);
	gtk_table_attach(GTK_TABLE(table), scale[0],
	    0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 2);
	gtk_table_attach(GTK_TABLE(table), scale[1],
	    0, 1, 4, 5, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 2);
	gtk_table_attach(GTK_TABLE(table), scale[2],
	    0, 1, 10, 11, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 2);

	/* add combo box */
	combo = gtk_combo_new();
	/*
	 * the following sets the active text to the current selected.
	 * I've tried a number of better ways, which fail on this
	 * version of GTK.
	 */
	switch (c_pallet) {
		case C_ORIGINAL:
			glist = g_list_append(glist, "original");
			glist = g_list_append(glist, "greyscale");
			glist = g_list_append(glist, "scale12");
			break;
		case C_GREYSCALE:
			glist = g_list_append(glist, "greyscale");
			glist = g_list_append(glist, "original");
			glist = g_list_append(glist, "scale12");
			break;
		case C_SCALE12:
			glist = g_list_append(glist, "scale12");
			glist = g_list_append(glist, "original");
			glist = g_list_append(glist, "greyscale");
			break;
	}
	gtk_combo_set_popdown_strings(GTK_COMBO(combo), glist);
	gtk_table_attach(GTK_TABLE(table), combo,
	    0, 1, 7, 8, GTK_SHRINK, GTK_SHRINK, 5, 2);

	/* add horizontal seperators */
	seperator[0] = gtk_hseparator_new();
	seperator[1] = gtk_hseparator_new();
	seperator[2] = gtk_hseparator_new();
	gtk_table_attach(GTK_TABLE(table), seperator[0],
	    0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 10);
	gtk_table_attach(GTK_TABLE(table), seperator[1],
	    0, 1, 5, 6, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 10);
	gtk_table_attach(GTK_TABLE(table), seperator[2],
	    0, 1, 8, 9, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 10);

	/* show */
	gtk_widget_show_all(dialog);

	/* process dialog */
	do {
		running = 0;
		result = gtk_dialog_run(GTK_DIALOG(dialog));
		switch (result) {
			case GTK_RESPONSE_APPLY:
				running = 1;
				/* FALLTHROUGH */
			case GTK_RESPONSE_ACCEPT:

				/* set globals */
				dt_resolution =
				    gtk_range_get_value(GTK_RANGE(scale[0]));
				dt_colour_factor =
				    gtk_range_get_value(GTK_RANGE(scale[1]));
				dt_update =
				    gtk_range_get_value(GTK_RANGE(scale[2]));
				menu_set_pallet((gchar *)gtk_entry_get_text(
				    GTK_ENTRY(GTK_COMBO(combo)->entry)));

				/* set line width as appropriate */
				if (dt_resolution < 4) {
					gdk_gc_set_line_attributes(c_gc, 1,
					    GDK_LINE_SOLID, 0, 0);
				} else {
					gdk_gc_set_line_attributes(c_gc, 2,
					    GDK_LINE_SOLID, 0, 0);
				}

				/* update screen status */
				update_options_info();
				dt_plot_resized = TRUE;
				dt_plot_exposed = TRUE;

				break;
			case GTK_RESPONSE_HELP:
				menu_options_help(GTK_WINDOW_TOPLEVEL);
				running = 1;
				break;
			default:
				break;
		}
	} while (running);

	/* destroy box */
	gtk_widget_destroy(dialog);
}

void
menu_save_dialog(gpointer data, guint action, GtkWidget *menuitem)
{
	GtkWidget	*dialog;
	char		pwd[LINE_LEN];
	char		*filename;

	/* create dialog */
	dialog = gtk_file_chooser_dialog_new("Save File",
	    GTK_WINDOW_TOPLEVEL,
	    GTK_FILE_CHOOSER_ACTION_SAVE,
	    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
	    NULL);

	/* fetch defaults */
	if (getcwd(pwd, LINE_LEN) == NULL)
		strcpy(pwd, "/tmp");
	if (lastname[0] == '\0')
		strcpy(lastname, "untitled.dtaz");

	/* set defaults */
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), pwd);
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), lastname);

	/* actually run the dialog now */
	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		filename = gtk_file_chooser_get_filename(
		    GTK_FILE_CHOOSER(dialog));

		/* save */
		menu_save_file(filename);
		strcpy(lastname, basename(filename));
		g_free(filename);
	}

	gtk_widget_destroy(dialog);
	g_free(pwd);
}

void
menu_save_file(char *filename)
{
	int		plot_max_x = dt_plot_da->allocation.width;
	int		plot_max_y = dt_plot_da->allocation.height;
	int		seek_max_x = dt_seek_da->allocation.width;
	int		seek_max_y = dt_seek_da->allocation.height;
	char		buf[sizeof (menu_save_hdr_t)];
	struct stat	filestat;
	int		i;
	int		j;
	int		out;
	int		ok;
	gint		result;

	menu_save_hdr_t	*hdrp;
	hdrp = (menu_save_hdr_t *)buf;

	/* prepare save file header */
	strcpy(hdrp->header, "DTraceTazTool");
	hdrp->endian = 1;
	hdrp->version = DT_SAVE_VERSION;
	hdrp->resolution = dt_resolution;
	hdrp->colour_factor = dt_colour_factor;
	hdrp->update = dt_update;
	hdrp->time = dt_time;
	hdrp->current_x = dt_current_x;
	hdrp->plot_max_x = plot_max_x;
	hdrp->plot_max_y = plot_max_y;
	hdrp->seek_max_x = seek_max_x;
	hdrp->seek_max_y = seek_max_y;
	hdrp->plot_max_raw = dt_plot_max_raw;
	hdrp->script = dt_script;
	hdrp->scale = dt_scale;
	hdrp->pallet = c_pallet;

	/* check for file existance */
	ok = 1;
	if (stat(filename, &filestat) == 0) {
		GtkWidget	*dialog;
		GtkWidget	*label;

		/* create confirmation box */
		dialog = gtk_dialog_new_with_buttons("Confirm Replace",
		    GTK_WINDOW_TOPLEVEL,
		    GTK_DIALOG_DESTROY_WITH_PARENT,
		    GTK_STOCK_OK,
		    GTK_RESPONSE_ACCEPT,
		    GTK_STOCK_CANCEL,
		    GTK_RESPONSE_REJECT,
		    NULL);

		/* set label */
		label = gtk_label_new(
		    "WARNING: File already exists. Replace?");
		gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
		    label);

		/* show */
		gtk_widget_show_all(dialog);

		/* run */
		result = gtk_dialog_run(GTK_DIALOG(dialog));
		switch (result) {
			case GTK_RESPONSE_ACCEPT:
				ok = 1;
				break;
			default:
				ok = 0;
				break;
		}

		gtk_widget_destroy(dialog);
	}

	/* quit if cancelled */
	if (ok == 0)
		return;

	/* write to file */
	if ((out = open(filename, O_WRONLY | O_CREAT, 0644)) == -1) {
		sb_err(dt_status, "Couldn't open file.");
		return;
	}

	/* write header */
	if (write(out, buf, sizeof (menu_save_hdr_t)) == -1) {
		(void) close(out);
		sb_err(dt_status, "Writing file.");
		return;
	}

	/* write data */
	for (j = 0; j < plot_max_x / dt_resolution; j++)
		write(out, (void *)&dt_plot_canvas[j][0],
		    sizeof (uint16_t) * DT_PLOT_MAX_Y);
	for (i = 0; i < plot_max_x / dt_resolution; i++) {
		write(out, &dt_seek_canvas[i].avg, sizeof (uint32_t));
		write(out, &dt_seek_canvas[i].max, sizeof (uint32_t));
	}
	/* BDG - add error checking */

	(void) close(out);
}

void
menu_open_dialog(gpointer data, guint action, GtkWidget *menuitem)
{
	GtkWidget	*dialog;

	/* create dialog */
	dialog = gtk_file_chooser_dialog_new("Open File",
	    GTK_WINDOW_TOPLEVEL,
	    GTK_FILE_CHOOSER_ACTION_OPEN,
	    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	    GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
	    NULL);

	/* actually run the dialog now */
	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		char *filename;
		filename = gtk_file_chooser_get_filename(
		    GTK_FILE_CHOOSER(dialog));

		/* open */
		menu_open_file(filename);

		g_free(filename);
	}

	gtk_widget_destroy(dialog);
}

void
menu_open_file(char *file)
{
	char		buf[sizeof (menu_save_hdr_t)];
	char		bbuf[LINE_LEN];
	menu_save_hdr_t	*hdrp;
	int		plot_max_x;
	int		plot_max_y;
	int		seek_max_x;
	int		seek_max_y;
	int		i;
	int		j;
	int		in;

	/* read from a file */
	if ((in = open(file, O_RDONLY)) == -1) {
		sb_err(dt_status, "Couldn't open file.");
		return;
	}

	/* read header */
	if (read(in, buf, sizeof (menu_save_hdr_t)) == -1) {
		sb_err(dt_status, "Couldn't read from file.");
		(void) close(in);
		return;
	}

	/* check this is of the correct type */
	hdrp = (menu_save_hdr_t *)buf;
	if (strcmp(hdrp->header, "DTraceTazTool") != 0) {
		sb_err(dt_status, "File is not a DTraceTazTool save.");
		(void) close(in);
		return;
	}

	/* clear data */
	plot_new_screen();

	/* process header */
	dt_resolution = hdrp->resolution;
	dt_colour_factor = hdrp->colour_factor;
	dt_update = hdrp->update;
	dt_time = hdrp->time;
	dt_current_x = hdrp->current_x;
	plot_max_x = hdrp->plot_max_x;
	plot_max_y = hdrp->plot_max_y;
	seek_max_x = hdrp->seek_max_x;
	seek_max_y = hdrp->seek_max_y;
	dt_plot_max_raw = hdrp->plot_max_raw;
	dt_script = hdrp->script;
	dt_scale = hdrp->scale;
	c_pallet = hdrp->pallet;

	/* read data */
	for (j = 0; j < plot_max_x / dt_resolution; j++)
		read(in, (void *)&dt_plot_canvas[j][0],
		    sizeof (uint16_t) * DT_PLOT_MAX_Y);
	for (i = 0; i < plot_max_x / dt_resolution; i++) {
		read(in, &dt_seek_canvas[i].avg, sizeof (uint32_t));
		read(in, &dt_seek_canvas[i].max, sizeof (uint32_t));
	}
	/* BDG - add error checking */

	(void) close(in);

	/* update screen status */
	update_options_info();
	sprintf(bbuf, "Block\n%llu", (u_longlong_t)dt_plot_max_raw);
	gtk_label_set_text(GTK_LABEL(dt_plot_label[2]), bbuf);
	gtk_widget_set_size_request(dt_plot_da, plot_max_x, plot_max_y);
	gtk_widget_set_size_request(dt_seek_da, seek_max_x, seek_max_y);
	dt_plot_resized = TRUE;
	dt_plot_exposed = TRUE;
}
